package cz.zweistein.df.soundsense.config;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.sun.org.apache.xerces.internal.parsers.DOMParser;

import cz.zweistein.df.soundsense.util.log.LoggerSource;

public class SoundsXML {
	private static Logger logger = LoggerSource.logger;
	
	private List<Sound> sounds;
	private String rootDirectory;
	
	public SoundsXML(String directory, boolean ignoreEmptySounds) {
		this.rootDirectory = directory;
		this.sounds = new LinkedList<Sound>();
		
		this.loadDirectory(directory, ignoreEmptySounds);
	}

	public SoundsXML(String directory) {
		this(directory, true);
	}
	
	private void loadDirectory(String directory, boolean ignoreEmptySounds) {
		logger.fine("Scanning directory '"+directory+"'.");
		File dir = new File(directory);
		
		String[] files = dir.list();
		
		if (files == null) {
			logger.info("'"+directory+"' is empty or invalid? Ignoring.");
		} else {
			Arrays.sort(files);
			for (int i = 0; i < files.length; i++) {
				
				String fileName = files[i];
				
				if (new File(directory+fileName).isDirectory()) {
					this.loadDirectory(directory+fileName+"/", ignoreEmptySounds);
				} else if (fileName.endsWith(".xml")) {
				
					try {
						logger.info("Loading config "+directory+fileName);
						this.loadFile(directory+fileName, ignoreEmptySounds);
					} catch (Exception e) {
						logger.severe("Failed to load "+ fileName + ": " + e.toString());
					}
				
				} else {
					logger.finest("'"+fileName+"' is not configuration file.");
				}
			}
		}
		
	}
	
	private void loadFile(String fileName, boolean ignoreEmptySounds) throws SAXException, IOException {
		
		Document doc = null;
		InputSource source = new InputSource(new FileInputStream(fileName));
		DOMParser parser = new DOMParser();
		parser.parse(source);
		doc = parser.getDocument();
		
		NodeList soundTags = doc.getElementsByTagName("sound");
		parseSounds(soundTags, fileName, ignoreEmptySounds);
		
	}

	
	private String parseStringAtribute(Node node, String atribute) {
		String string = null;
		Node stringNode = node.getAttributes().getNamedItem(atribute);
		if (stringNode != null) {
			string = stringNode.getNodeValue();
		}
		return string;
	}
	
	private Long parseIntegerAtribute(Node node, String atribute) {
		Long value = null;
		if (node.getAttributes().getNamedItem(atribute) != null) {
			String numberText = node.getAttributes().getNamedItem(atribute).getNodeValue();
			try {
				value = Long.parseLong(numberText);
			} catch (NumberFormatException e) {
				logger.info(atribute+" '"+numberText+"' is not recognized as a number.");
			}
		}
		return value;
	}

	private void parseSounds(NodeList soundTags, String fileName, boolean ignoreEmptySounds) {
		for (int i = 0; i < soundTags.getLength(); i++) {
			Node soundNode = soundTags.item(i);
			
			String logPattern = parseStringAtribute(soundNode, "logPattern");
			
			if (logPattern == null) {
				logger.info("Sound tag without 'logPattern' attribute encountered.");
			}
			
			String channel = null;
			if (soundNode.getAttributes().getNamedItem("channel") != null) {
				channel = soundNode.getAttributes().getNamedItem("channel").getNodeValue();
			}
			
			Loop channelLoop = null;
			if (soundNode.getAttributes().getNamedItem("loop") != null) {
				String loop = soundNode.getAttributes().getNamedItem("loop").getNodeValue();
				if ("true".equals(loop) || "start".equals(loop)) {
					channelLoop = Loop.START_LOOPING;
				} else if ("stop".equals(loop) ) {
					channelLoop = Loop.STOP_LOOPING;
				} else {
					logger.fine("Loop parameter '"+loop+"' not recognized.");
				}
			}
			
			boolean haltOnMatch = true;
			if (soundNode.getAttributes().getNamedItem("haltOnMatch") != null) {
				String haltOnMatchText = soundNode.getAttributes().getNamedItem("haltOnMatch").getNodeValue();
				if ("true".equals(haltOnMatchText)) {
					haltOnMatch = true;
				} else if ("false".equals(haltOnMatchText)) {
					haltOnMatch = false;
				} else {
					logger.fine("HaltOnMatch parameter '"+haltOnMatchText+"' not recognized, using default "+haltOnMatch+".");
				}
			}
			
			Long delay = parseIntegerAtribute(soundNode, "delay");

			Long concurency = parseIntegerAtribute(soundNode, "concurency");
			
			Long timeout = parseIntegerAtribute(soundNode, "timeout");

			Long propability = parseIntegerAtribute(soundNode, "propability");
			
			List<SoundFile> soundFiles = new ArrayList<SoundFile>();
			
			NodeList soundElements = soundNode.getChildNodes();
			for (int j = 0; j < soundElements.getLength(); j++) {
				Node configNode = soundElements.item(j);
				String name = configNode.getLocalName();
				
				if ("soundFile".equals(name)) {
					SoundFile soundFile = this.parseSoundFile(configNode, fileName);
					if (soundFile != null) {
						soundFiles.add(soundFile);
					}
				}
			}
			
			if (logPattern == null || logPattern.length() == 0) {
				logger.info("Sound does not have logPattern, ignoring.");
			} else if ((soundFiles.size() == 0 && ignoreEmptySounds) && (channelLoop != Loop.STOP_LOOPING)) {
				logger.fine("Sound for '"+logPattern+"' does not contain any soundFiles and is not speech or end of loop, ignoring.");
			} else {
				this.sounds.add(new Sound(fileName, soundFiles, logPattern, channelLoop, channel, concurency, haltOnMatch, timeout, delay, propability));
				logger.finest("Added sound for "+logPattern);
			}
		}

	}

	private SoundFile parseSoundFile(Node configNode, String fileName) {
		
		Node fileNameAttribute = configNode.getAttributes().getNamedItem("fileName");
		if (fileNameAttribute == null) {
			logger.info("Could not locate 'fileName' attribute of a sound element, please check spelling.");
			return null;
		}
		
		String soundFile = new File(fileName).getParent()+"/"+fileNameAttribute.getNodeValue();
		
		int weight = 100;
		if (configNode.getAttributes().getNamedItem("weight") != null) {
			String weightText = configNode.getAttributes().getNamedItem("weight").getNodeValue();
			try {
				weight = Integer.parseInt(weightText);
			} catch (NumberFormatException e) {
				logger.info("Weight '"+weightText+"' for '"+soundFile+"' is not recognized as a number, using default "+weight+".");
			}
		}
		
		float volumeAdjustment = 0;
		if (configNode.getAttributes().getNamedItem("volumeAdjustment") != null) {
			String volumeAdjustmentText = configNode.getAttributes().getNamedItem("volumeAdjustment").getNodeValue();
			try {
				volumeAdjustment = Float.parseFloat(volumeAdjustmentText);
			} catch (NumberFormatException e) {
				logger.info("VolumeAdjustment '"+volumeAdjustmentText+"' for '"+soundFile+"' is not recognized as a number, using default "+volumeAdjustment+".");
			}
		}
		
		if (new File(soundFile).exists()) {
			return new SoundFile(soundFile, weight, volumeAdjustment);
		} else {
			logger.info("Did not find "+soundFile+", ignoring.");
			return null;
		}
	}

	public List<Sound> getSounds() {
		return this.sounds;
	}

	public List<String> getXMLFiles() {
		List<String> xmlFiles = new ArrayList<String>();
		for (Sound sound: this.sounds) {
			if (!xmlFiles.contains(sound.getParentFile())) {
				xmlFiles.add(sound.getParentFile());
			}
		}
		return xmlFiles;
	}
	
	public List<Sound> getSoundsByXMLFile(String xmlFile) {
		List<Sound> sounds = new ArrayList<Sound>();
		for (Sound sound: this.sounds) {
			if (xmlFile.equals(sound.getParentFile())) {
				sounds.add(sound);
			}
		}
		return sounds;
	}

	@Override
	public String toString() {
		return this.rootDirectory;
	}
	
}
